/*
 * $Id: at.c,v 1.1.2.8 2006/12/06 07:47:09 bobyu Exp $
 */
 
/*******************************************************
Address Translation Table MIB
********************************************************/
#include <config.h>

/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include <ucd-snmp/mibgroup/mibII/group_mib.h>
#include "mibgroup/mibII/at.h"

static group_parameters_t *atmib_parms = NULL;

/* Path to this MIB wrt to the global MIB tree */
oid at_variables_oid[] = { 1,3,6,1,2,1,3 };

/* Variables that this MIB supports */
struct variable4 at_variables[] = {
#define ATIFINDEX       1
    {ATIFINDEX,  ASN_INTEGER,  RWRITE,  var_atTable, 3,  { 1, 1, 1 }},
#define ATPHYSADDRESS       2
    {ATPHYSADDRESS,  ASN_OCTET_STR,  RWRITE,  var_atTable, 3,  { 1, 1, 2 }},
#define ATNETADDRESS        3
    {ATNETADDRESS,  ASN_IPADDRESS,  RWRITE,  var_atTable, 3,  { 1, 1, 3 }},
};

/** Initializes the at module **/
void
init_at(void)
{
    AGENT_ENTER;

    /* BRCM SNMP init */
    atmib_parms = (group_parameters_t *) \
                sal_alloc(sizeof(group_parameters_t), MIBII_AT_MIB);

    if (atmib_parms != NULL) {
        atmib_parms->group = GROUP_AT;

        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB(MIBII_AT_MIB, at_variables, variable4,
                   at_variables_oid);
    }

    AGENT_EXIT;
}

/*
 * var_at():
 *   This function is called every time the agent gets a request for
 *   a scalar variable
 */
unsigned char *
var_at(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    AGENT_ENTER;

    if (header_generic(vp,name,length,exact,var_len,write_method)
                                                  == MATCH_FAILED ) {
        return NULL;
    }

    switch(vp->magic) {
        default:
            ERROR_MSG("");
    }

    AGENT_EXIT;
    return NULL;
}

static int
header_atTable(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define AT_NAME_LENGTH    10
#define AT_TOTAL_LENGTH   15
    AGENT_ENTER;
    oid newname[MAX_OID_LEN];
    int i, result, found = FALSE;
    int j = 0;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));
    /* Now, get the entry by index and validate request */
    for (i = 1; i <= 5; i++) {
        newname[AT_NAME_LENGTH] = (oid)i;

        for (j = 0; j < IP_ADDRESS_LEN; j++)  { 
            newname[AT_NAME_LENGTH+j+1] = (oid)0x0a;
        }
        result = snmp_oid_compare(name, *length, newname, 
                                  (int)vp->namelen + 1);
        if ((exact && (result == 0)) || (!exact && (result < 0))) {
            found = TRUE;
            break;
        }
    }

    if (!found) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy((char *)name,(char *)newname, AT_TOTAL_LENGTH * sizeof(oid));
    *length = AT_TOTAL_LENGTH;
    *var_len = sizeof(long);

    AGENT_EXIT;
    return MATCH_SUCCEEDED;
}


/*
 * var_atTable():
 *   Handle this table separately from the scalar value case.
 *   The workings of this are basically the same as for var_at above.
 */
unsigned char *
var_atTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    /* variables we may use later */
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];

    /* TODO: Remove MIB once implemented. Unimplemented MIB */
    return NULL;

    /* Simple table implementation */
    if (header_atTable(vp,name,length,exact,var_len,write_method)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case ATIFINDEX:
            *write_method = write_atIfIndex;
            /* atmib_parms->group_obj = AT_IFINDEX; */
            long_ret = (int) name[10];
            return (unsigned char *) &long_ret;
            break;

        case ATPHYSADDRESS:
            *write_method = write_atPhysAddress;
            atmib_parms->group_obj = AT_PHYSADDRESS;
            break;
        case ATNETADDRESS:
            *write_method = write_atNetAddress;
            atmib_parms->group_obj = AT_NETADDRESS;
            break;
        default:
            ERROR_MSG("Undefined AT MIB Reference");
    }

    atmib_parms->callback_rw = CALLBACK_ACT_READ;
    atmib_parms->group = GROUP_AT; 

    /* Callback */
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_GROUPS,
                        atmib_parms);

    /* Return results */
    switch(vp->magic) {
        case ATPHYSADDRESS:
            *var_len = MAC_ADDRESS_LEN;
            memcpy (string, atmib_parms->buf, *var_len);
            break;

        case ATNETADDRESS:
            string[0] = atmib_parms->buf[0];
            string[1] = atmib_parms->buf[1];
            string[2] = atmib_parms->buf[2];
            string[3] = atmib_parms->buf[3];
            *var_len = IP_ADDRESS_LEN;
            break;
    }

    AGENT_EXIT;
    return (unsigned char *) string;
}


int
write_atIfIndex(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    long value = 0;

    if (var_val) {
        value = * (long *) var_val;
    }

    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to at not ASN_INTEGER\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to at: bad length\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGLENGTH;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
          break;

        case ACTION:
          break;

        case UNDO:
          break;

        case COMMIT:
          break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

int
write_atPhysAddress(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    int size;

    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_OCTET_STR) {
              fprintf(stderr, "write to at not ASN_OCTET_STR\n");
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(char)) {
              fprintf(stderr,"write to at: bad length\n");
              return SNMP_ERR_WRONGLENGTH;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
          break;

        case ACTION:
          break;

        case UNDO:
          break;

        case COMMIT:
          break;
    }
    return SNMP_ERR_NOERROR;
}

int
write_atNetAddress(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    char value;
    int size;

    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_COUNTER) {
              fprintf(stderr, "write to at not ASN_COUNTER\n");
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(int)) {
              fprintf(stderr,"write to at: bad length\n");
              return SNMP_ERR_WRONGLENGTH;
          }
          break;

        case RESERVE2:
          size  = var_val_len;
          value = * (int *) var_val;
          break;

        case FREE:
          break;

        case ACTION:
          break;

        case UNDO:
          break;

        case COMMIT:
          break;
    }
    return SNMP_ERR_NOERROR;
}
