/*
 * $Id: dot1dPriority.c,v 1.1.2.7 2006/11/06 04:17:20 bobyu Exp $
 */

/* This should always be included first before anything else */
#include <config.h>

/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include "mibgroup/mibII/dot1dPriority.h"
#include <ucd-snmp/mibgroup/mibII/bridge_mib.h>

#define C_MAX_TRAFFIC_CLASSES  8

/* Argument to be passed with callback function */
static bridge_parameters_t *dot1dprio_parms = NULL;

/* Path to this MIB group */
oid dot1dPriority_variables_oid[] = { 1,3,6,1,2,1,17,6,1,2 };

/* Variables that this MIB supports */
struct variable4 dot1dPriority_variables[] = {
/*  magic number        , variable type , ro/rw , callback fn  , L, oidsuffix */

#define DOT1DPORTDEFAULTUSERPRIORITY        1
    {DOT1DPORTDEFAULTUSERPRIORITY,  ASN_INTEGER,  RWRITE,  var_dot1dPortPriorityTable, 3,  { 1, 1, 1 }},
#define DOT1DPORTNUMTRAFFICCLASSES      2
    {DOT1DPORTNUMTRAFFICCLASSES,  ASN_INTEGER,  RWRITE,  var_dot1dPortPriorityTable, 3,  { 1, 1, 2 }},

#ifdef DOT1P_ENHANCE 
    #define DOT1DUSERPRIORITY       3
        {DOT1DUSERPRIORITY,  ASN_INTEGER,  RONLY,   var_dot1dUserPriorityRegenTable, 3,  { 2, 1, 1 }},
    #define DOT1DREGENUSERPRIORITY      4
        {DOT1DREGENUSERPRIORITY,  ASN_INTEGER,  RWRITE,  var_dot1dUserPriorityRegenTable, 3,  { 2, 1, 2 }},
    #define DOT1DPORTOUTBOUNDACCESSPRIORITY     7
        {DOT1DPORTOUTBOUNDACCESSPRIORITY,  ASN_INTEGER,  RONLY,   var_dot1dPortOutboundAccessPriorityTable, 3,  { 4, 1, 1 }},
#endif

#define DOT1DTRAFFICCLASSPRIORITY       5
    {DOT1DTRAFFICCLASSPRIORITY,  ASN_INTEGER,  RONLY,   var_dot1dTrafficClassTable, 3,  { 3, 1, 1 }},
#define DOT1DTRAFFICCLASS       6
    {DOT1DTRAFFICCLASS,  ASN_INTEGER,  RWRITE,  var_dot1dTrafficClassTable, 3,  { 3, 1, 2 }},
};

/** Initializes the dot1dPriority module */
void
init_dot1dPriority(void)
{
    AGENT_ENTER;

    /* BRCM SNMP init */
    dot1dprio_parms = (bridge_parameters_t *) \
                sal_alloc(sizeof(bridge_parameters_t), MIBII_DOT1DPRIORITY_MIB);

    if (dot1dprio_parms) {
        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB(MIBII_DOT1DPRIORITY_MIB, dot1dPriority_variables, variable4,
                   dot1dPriority_variables_oid);

        dot1dprio_parms->group = BRIDGE_DOT1DPRIORITY;
    }

    AGENT_EXIT;
}

/*
 * var_dot1dPriority(): Top level function.
 */
unsigned char *
var_dot1dPriority(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    AGENT_ENTER;

    if (header_generic(vp,name,length,exact,var_len,write_method)
                                                      == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    AGENT_EXIT;
    return NULL;
}

static int
header_dot1dPortPriorityEntry(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1DPORTPRIO_ENTRY_LENGTH  13
    AGENT_ENTER;
    oid newname[MAX_OID_LEN];
    int i, result;
    int port_number;
    int found = FALSE;

    /* copy the query OID */
    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));

    /* Get the OID requested */
    for(i=1; i <= interface_count; i++)  {
        newname[DOT1DPORTPRIO_ENTRY_LENGTH] = (oid)i;
        result = snmp_oid_compare(
                name, *length, newname, (int)vp->namelen + 1);
        if ((exact && (result == 0)) || (!exact && (result < 0))) {
            found = TRUE;
            port_number = i;
            break;
        }
    }

    /* TODO: Take AGG ports into consideration? */

    if ( found == FALSE ) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 1) * sizeof(oid));
    *length = vp->namelen + 1;
    *var_len = sizeof(long);

    AGENT_EXIT;
    return port_number;
}

/*
 * var_dot1dPortPriorityTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1dPriority above.
 */
unsigned char *
var_dot1dPortPriorityTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;

    /* Call custom validation function */
    snmp_mib_setInterfaceCount();
    if ((dot1dprio_parms->index =
                    header_dot1dPortPriorityEntry(vp,name,length,exact,
                                var_len,write_method)) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    AGENT_DBG(("1D priority index is %d", dot1dprio_parms->index));
    switch(vp->magic) {
        case DOT1DPORTDEFAULTUSERPRIORITY:
            *write_method = write_dot1dPortDefaultUserPriority;
            dot1dprio_parms->group_obj = DOT1D_PORTDEFAULTUSERPRIORITY;
            break;

        case DOT1DPORTNUMTRAFFICCLASSES:
            *write_method = write_dot1dPortNumTrafficClasses;
            dot1dprio_parms->group_obj = DOT1D_PORTNUMTRAFFICCLASSES;
            break;

        default:
            ERROR_MSG("Undefined MIB reference in dot1dPortPriority");
            AGENT_EXIT;
            return NULL;
    }

    /* Callback */
    dot1dprio_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dprio_parms);
    
    /* Give the value back */
    long_ret = (long)dot1dprio_parms->value;
    AGENT_EXIT;
    return (u_char *)&long_ret;
}

/*
 * Write user priority values
 */
int
write_dot1dPortDefaultUserPriority(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    int value = 0;

    if (var_val) {
        value = * (int *) var_val;
        AGENT_DBG(("Setting value %d", value));
    }

    dot1dprio_parms->index = (int) name[DOT1DPORTPRIO_ENTRY_LENGTH];
    AGENT_DBG(("Index is %d", dot1dprio_parms->index));

    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to dot1dPriority not ASN_INTEGER\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGTYPE;
          }

          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to dot1dPriority: bad length\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGLENGTH;
          }

          if (value < 0 || value > (SS_QOS_MAX_PRIORITY - 1)) {
              fprintf(stderr,"write to dot1dPriority: bad value\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGVALUE;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
             /* Release any resources that have been allocated */
          break;

        case ACTION:
          break;

        case UNDO:
             /* Back out any changes made in the ACTION case */
          break;

        case COMMIT:
            dot1dprio_parms->value = value;
            dot1dprio_parms->group_obj = DOT1D_PORTDEFAULTUSERPRIORITY;
            dot1dprio_parms->callback_rw = CALLBACK_ACT_WRITE;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                                SNMPD_CALLBACK_BRIDGE,
                                dot1dprio_parms);
            if (dot1dprio_parms->rv == MIB_STATUS_ERROR) {
                AGENT_EXIT;
                return SNMP_ERR_COMMITFAILED;
            }
          break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

/*
 * Write traffic class values
 */
int
write_dot1dPortNumTrafficClasses(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    int value = 0;

    if (var_val) {
        value = * (int *) var_val;
        AGENT_DBG(("Num traff class = %d", value));
    }

    dot1dprio_parms->index = (int) name[DOT1DPORTPRIO_ENTRY_LENGTH];
    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to dot1dPriority not ASN_INTEGER\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to dot1dPriority: bad length\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGLENGTH;
          }

          /* SDK def says we can have 1,2,4 or 8 queues */
          if ((value < 1 || value > C_MAX_TRAFFIC_CLASSES) || 
                      (value != 1 && value != 2 && value != 4 && value != C_MAX_TRAFFIC_CLASSES)) {
              fprintf(stderr, "write to dot1dPriority: bad value\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGVALUE;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
             /* Release any resources that have been allocated */
          break;

        case ACTION:
          break;

        case UNDO:
             /* Back out any changes made in the ACTION case */
          break;

        case COMMIT:
            dot1dprio_parms->value = value;
            dot1dprio_parms->callback_rw = CALLBACK_ACT_WRITE;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                                SNMPD_CALLBACK_BRIDGE,
                                dot1dprio_parms);
            if (dot1dprio_parms->rv == MIB_STATUS_ERROR) {
                AGENT_EXIT;
                return SNMP_ERR_COMMITFAILED;
            }
          break;
    }
    
    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

#ifdef DOT1P_ENHANCE 
/*
 * user priority regen table
 * NOTE: This is not currently supported
 */
unsigned char *
var_dot1dUserPriorityRegenTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;

    /* Validate the table */
    if (header_simple_table(vp,name,length,exact,var_len,write_method, 
                                    8) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
    case DOT1DUSERPRIORITY:
        dot1dprio_parms->group_obj = DOT1D_USERPRIORITY;
        break;
        
    case DOT1DREGENUSERPRIORITY:
        *write_method = write_dot1dRegenUserPriority;
        dot1dprio_parms->group_obj = DOT1D_REGENUSERPRIORITY;
        break;
    default:
        ERROR_MSG("Undefined MIB reference");
        return NULL;
    }

    dot1dprio_parms->callback_rw = CALLBACK_ACT_READ;
    dot1dprio_parms->group = BRIDGE_DOT1DPRIORITY; 

    /* Callback */
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dprio_parms);
    
    /* Give the value back */
    long_ret = (long)dot1dprio_parms->value;
    return (u_char *)&long_ret;
}

int
write_dot1dRegenUserPriority(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    int value = 0;

    if (var_val) {
        value  = * (int *) var_val;
    }

    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to dot1dPriority not ASN_INTEGER\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to dot1dPriority: bad length\n");
              AGENT_EXIT;
              return SNMP_ERR_WRONGLENGTH;
          }
          if (value < 0 || value > 7) {
              AGENT_EXIT;
              return SNMP_ERR_WRONGVALUE;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
             /* Release any resources that have been allocated */
          break;

        case ACTION:
             /*
              * The variable has been stored in 'value' for you to use,
              * and you have just been asked to do something with it.
              * Note that anything done here must be reversable in the UNDO case
              */
          break;

        case UNDO:
             /* Back out any changes made in the ACTION case */
          break;

        case COMMIT:
          break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}
#endif /* DOT1P_ENHANCE  */

/*
 * Build valid OID, with port number and traffic class
 */
static int
header_dot1dTrafficClassEntry(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1DTRAFFIC_CLASS_NAME_LEN 13
    AGENT_ENTER;
    oid newname[MAX_OID_LEN];
    int i, j, result;
    int port_number;
    int found = FALSE;
    int offset;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));

    for(i = 1; i <= interface_count && found == FALSE; i++) {
        newname[DOT1DTRAFFIC_CLASS_NAME_LEN] = (oid)i;
        /* Iterate for all traffic classes */
        for(j = 0; j < C_MAX_TRAFFIC_CLASSES; j++) {
            newname[DOT1DTRAFFIC_CLASS_NAME_LEN + 1] = (oid) j;
            result = snmp_oid_compare(
                    name, *length, newname, (int)vp->namelen + 2);
            if ((exact && (result == 0)) || (!exact && (result < 0))) {
                found = TRUE;
                port_number = i ;
				AGENT_DBG(("Found match at %d",i));
                break;
            }
        }
    }

    if ( found == FALSE ) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy( (char *)name,(char *)newname, ((int)vp->namelen + 2) * sizeof(oid));
    *length = vp->namelen + 2;
    *var_len = sizeof(long);

    AGENT_EXIT;
    return port_number;
}

/*
 * var_dot1dTrafficClassTable(): Priority to queue mapping.
 *  This function contains Multi-Index tables.
 *   The workings of this are basically the same as for var_dot1dPriority above.
 */
unsigned char *
var_dot1dTrafficClassTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;

    /* Custom validation function */
    snmp_mib_setInterfaceCount();
    if ((header_dot1dTrafficClassEntry(vp,name,length,exact,
                               var_len,write_method)) == MATCH_FAILED ) {
        return NULL;
    }

    dot1dprio_parms->index = name[DOT1DTRAFFIC_CLASS_NAME_LEN +1]; 
    switch(vp->magic) {
        case DOT1DTRAFFICCLASSPRIORITY:
            dot1dprio_parms->group_obj = DOT1D_TRAFFICCLASSPRIORITY;
            break;

        case DOT1DTRAFFICCLASS:
            *write_method = write_dot1dTrafficClass;
            dot1dprio_parms->group_obj = DOT1D_TRAFFICCLASS;
            break;

        default:
            ERROR_MSG("Undefined MIB access in var_dot1dTrafficClassTable");
    }

    /* Callback */
    dot1dprio_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dprio_parms);

    /* Give the value back */
    long_ret = (long)dot1dprio_parms->value;
    return (u_char *)&long_ret;
}

int
write_dot1dTrafficClass(int      action,
            u_char   *var_val,
            u_char   var_val_type,
            size_t   var_val_len,
            u_char   *statP,
            oid      *name,
            size_t   name_len)
{
    AGENT_ENTER;
    int value = 0;
    int num_qs;

    if (var_val) {
        value = * (int *) var_val;
		AGENT_DBG(("Value = %d", value));
    }

    /* Get number of CosQs on the system */
    dot1dprio_parms->group_obj = DOT1D_PORTNUMTRAFFICCLASSES;
    dot1dprio_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dprio_parms);
    num_qs = dot1dprio_parms->value;

    dot1dprio_parms->group_obj = DOT1D_TRAFFICCLASS;
    dot1dprio_parms->index = name[DOT1DTRAFFIC_CLASS_NAME_LEN +1]; 
    switch ( action ) {
        case RESERVE1:
          if (var_val_type != ASN_INTEGER) {
              fprintf(stderr, "write to dot1dPriority not ASN_INTEGER\n");
              return SNMP_ERR_WRONGTYPE;
          }
          if (var_val_len > sizeof(long)) {
              fprintf(stderr,"write to dot1dPriority: bad length\n");
              return SNMP_ERR_WRONGLENGTH;
          }
          
          if (value < 0 || value > num_qs - 1) {
              fprintf(stderr, "write to dot1dPriority: bad value\n");
              return SNMP_ERR_WRONGVALUE;
          }
          break;

        case RESERVE2:
          break;

        case FREE:
             /* Release any resources that have been allocated */
          break;

        case ACTION:
          break;

        case UNDO:
             /* Back out any changes made in the ACTION case */
          break;

        case COMMIT:
            dot1dprio_parms->value = value;
            dot1dprio_parms->callback_rw = CALLBACK_ACT_WRITE;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                                SNMPD_CALLBACK_BRIDGE,
                                dot1dprio_parms);
            if (dot1dprio_parms->rv == MIB_STATUS_ERROR) {
                AGENT_EXIT;
                return SNMP_ERR_COMMITFAILED;
            }
          break;
    }
    return SNMP_ERR_NOERROR;
}

#ifdef DOT1P_ENHANCE 
/*
 * var_dot1dPortOutboundAccessPriorityTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1dPriority above.
 */
unsigned char *
var_dot1dPortOutboundAccessPriorityTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;

    /*
     * Simple table validation
     */
    if (header_simple_table(vp,name,length,exact,var_len,write_method, 10)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    switch(vp->magic) {
        case DOT1DPORTOUTBOUNDACCESSPRIORITY:
            dot1dprio_parms->group_obj = DOT1D_PORTOUTBOUNDACCESSPRIORITY;
            break;

        default:
            AGENT_EXIT;
            return NULL;
    }

    return NULL;
}

#endif /* DOT1P_ENHANCE  */

