
/*
 * $Id: dot1dStatic.c,v 1.1.2.10 2006/12/20 07:18:05 bobyu Exp $
 */

/* This should always be included first before anything else */
#include <config.h>

/* minimal include directives */
#include "mibincl.h"
#include "mibgroup/util_funcs.h"

#include "mibgroup/mibII/dot1dStatic.h"
#include <ucd-snmp/mibgroup/mibII/bridge_mib.h>

/* Argument to callback function */
static bridge_parameters_t *dot1dstatic_parms = NULL;

/* Index of l2 entry for which match was found */
static int static_entry_num = 0;

/* Path to this MIB WRT global MIB tree*/
oid dot1dStatic_variables_oid[] = { 1,3,6,1,2,1,17,5 };

/* Variables that this MIB supports */
struct variable4 dot1dStatic_variables[] = {
    #define DOT1DSTATICADDRESS      1
        {DOT1DSTATICADDRESS,  ASN_OCTET_STR,  RWRITE,  var_dot1dStaticTable, 3,  { 1, 1, 1 }},
    #define DOT1DSTATICRECEIVEPORT      2
        {DOT1DSTATICRECEIVEPORT,  ASN_INTEGER,  RWRITE,  var_dot1dStaticTable, 3,  { 1, 1, 2 }},
    #define DOT1DSTATICALLOWEDTOGOTO        3
        {DOT1DSTATICALLOWEDTOGOTO,  ASN_OCTET_STR,  RWRITE,  var_dot1dStaticTable, 3,  { 1, 1, 3 }},
    #define DOT1DSTATICSTATUS       4
        {DOT1DSTATICSTATUS,  ASN_INTEGER,  RWRITE,  var_dot1dStaticTable, 3,  { 1, 1, 4 }},
};

/** Initializes the dot1dStatic module */
void
init_dot1dStatic(void)
{
    AGENT_ENTER;
#if 0 /* No dot1dStatic support needed */    
    /* BRCM SNMP init */
    dot1dstatic_parms = (bridge_parameters_t *) \
                sal_alloc(sizeof(bridge_parameters_t), MIBII_DOT1DSTATIC_MIB);

    if (dot1dstatic_parms != NULL) {
        /* register ourselves with the agent to handle our mib tree */
        REGISTER_MIB(MIBII_DOT1DSTATIC_MIB, dot1dStatic_variables, variable4,
                      dot1dStatic_variables_oid);

        dot1dstatic_parms->group = BRIDGE_DOT1DSTATIC; 
    }
#endif
    AGENT_EXIT;
}

/*
 * var_dot1dStatic():
 */
unsigned char *
var_dot1dStatic(struct variable *vp, 
                oid     *name, 
                size_t  *length, 
                int     exact, 
                size_t  *var_len, 
                WriteMethod **write_method)
{
    AGENT_ENTER;
    
    if (header_generic(vp,name,length,exact,var_len,write_method)
                                                == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    return NULL;
}

/*
 * Validate the MAC Address
 */
static int
header_dot1dStaticEntry(struct variable *vp,
           oid *name,
           size_t *length,
           int exact,
           size_t *var_len,
           WriteMethod **write_method)
{
#define DOT1DSTATICENTRY_NAME_LENGTH    11
#define DOT1DSTATIC_TOTAL_OID_LENGTH 18
    AGENT_ENTER;
    oid newname[MAX_OID_LEN];
    oid lowest[MAX_OID_LEN];
    int i, result, found = FALSE;
    int num_mac_addr;
    unsigned char mac[6];
    int j = 0;

    memcpy( (char *)newname,(char *)vp->name, (int)vp->namelen * sizeof(oid));

    /* Set the entry number to an invalid index at the start */
    static_entry_num = -1;
    lowest[0] = 99999;

    /* Let us set the write method */
    *write_method = write_dot1dStaticEntry;

    /* dump l2 entries first */
    dot1dstatic_parms->group_obj = DOT1D_STATICADDRESSDUMP;
    dot1dstatic_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dstatic_parms);

    /* Get count of number of L2 entries */
    dot1dstatic_parms->group_obj = DOT1D_STATICADDRESSCOUNT;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dstatic_parms);
    num_mac_addr = dot1dstatic_parms->value;
    AGENT_DBG(("Num mac addr = %d", num_mac_addr));

    /* Now, get the entry by index and validate request */
    for (i = 0; i < num_mac_addr; i++) {
        dot1dstatic_parms->index = i;
        
        /* continue if static */
        dot1dstatic_parms->group_obj = DOT1D_STATICSTATUS;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dstatic_parms);
        if (dot1dstatic_parms->value != C_STATIC_STATUS_PERMANENT &&
                dot1dstatic_parms->value != C_STATIC_STATUS_DELONRESET) {
            continue;
        }

        /* get the mac address by index */
        dot1dstatic_parms->group_obj = DOT1D_STATICADDRESS;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dstatic_parms);
        memcpy (mac, dot1dstatic_parms->buf, 6);

        for (j = 0; j < MAC_ADDRESS_LEN; j++)  { /* 6 octets for MAC Address */
            newname[DOT1DSTATICENTRY_NAME_LENGTH+j] = (oid)mac[j];
        }

        /* get the port associated with the mac address */
        dot1dstatic_parms->group_obj = DOT1D_STATICRECEIVEPORT;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dstatic_parms);
        /* Port is 1 based */
        newname[DOT1DSTATICENTRY_NAME_LENGTH + MAC_ADDRESS_LEN] = 
                                        (oid)dot1dstatic_parms->value;
        AGENT_DBG(("Port Number %d", 
                              newname[DOT1DSTATICENTRY_NAME_LENGTH + 6]));

        /* if it's a SET/GET request, exact is true. So return only if there
         * is an exact match or snmp_oid_compare returns 0
         */
        if (exact) {
            if (snmp_oid_compare(name, *length, newname, 
                        DOT1DSTATIC_TOTAL_OID_LENGTH) == 0) {
                memcpy(lowest, newname, 
                                DOT1DSTATIC_TOTAL_OID_LENGTH * sizeof(oid));
                found = TRUE;
                static_entry_num = i;
				break;
            }
        } else {
            if (snmp_oid_compare(newname, DOT1DSTATIC_TOTAL_OID_LENGTH, 
                        name, *length) > 0 &&
                snmp_oid_compare(newname, DOT1DSTATIC_TOTAL_OID_LENGTH, lowest, 
                    DOT1DSTATIC_TOTAL_OID_LENGTH) < 0) {
                memcpy((char *) lowest, (char *) newname,
                       DOT1DSTATIC_TOTAL_OID_LENGTH * sizeof(oid));
                found = TRUE;
                static_entry_num = i;
            }
        }
    }

    if (!found) {
        AGENT_EXIT;
        return MATCH_FAILED;
    }

    memcpy((char *)name,(char *)lowest, DOT1DSTATIC_TOTAL_OID_LENGTH * 
                                                           sizeof(oid));
    *length = DOT1DSTATIC_TOTAL_OID_LENGTH;
    *var_len = sizeof(long);

    AGENT_EXIT;
    return MATCH_SUCCEEDED;
}

/*
 * var_dot1dStaticTable():
 *  This function contains Multi-Index tables
 *   The workings of this are basically the same as for var_dot1dStatic above.
 */
unsigned char *
var_dot1dStaticTable(struct variable *vp,
            oid     *name,
            size_t  *length,
            int     exact,
            size_t  *var_len,
            WriteMethod **write_method)
{
    AGENT_ENTER;
    static long long_ret;
    static unsigned char string[SPRINT_MAX_LEN];


    /* Simple table validation */
    snmp_mib_setInterfaceCount();
    if (header_dot1dStaticEntry(vp,name,length,exact,
                            var_len,write_method) == MATCH_FAILED ) {
        AGENT_EXIT;
        return NULL;
    }

    /* Set the index */
    dot1dstatic_parms->index = static_entry_num;
    switch(vp->magic) {
        case DOT1DSTATICADDRESS:
            dot1dstatic_parms->group_obj = DOT1D_STATICADDRESS;
            break;

        case DOT1DSTATICRECEIVEPORT:
            dot1dstatic_parms->group_obj = DOT1D_STATICRECEIVEPORT;
            break;

        case DOT1DSTATICALLOWEDTOGOTO:
            dot1dstatic_parms->group_obj = DOT1D_STATICALLOWEDTOGOTO;
            break;

        case DOT1DSTATICSTATUS:
            dot1dstatic_parms->group_obj = DOT1D_STATICSTATUS;
            break;
        default:
            ERROR_MSG("Undefined dot1dStatic MIB reference");
            AGENT_EXIT;
            return NULL;
    }

    /* Callback */
    dot1dstatic_parms->callback_rw = CALLBACK_ACT_READ;
    snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                        SNMPD_CALLBACK_BRIDGE,
                        dot1dstatic_parms);

    /* Give the value back */
    switch(vp->magic) {
        case DOT1DSTATICADDRESS:
            *var_len = MAC_ADDRESS_LEN;
            memcpy (string, dot1dstatic_parms->buf, *var_len);
            return (unsigned char *) string;

        case DOT1DSTATICALLOWEDTOGOTO:
		    /* Allowed to go to all ports */
            *var_len = (interface_count/INTERFACE_BYTE_COUNT);
            memcpy (string, dot1dstatic_parms->buf, *var_len);
            return (unsigned char *) string;
    }

    long_ret = (long)dot1dstatic_parms->value;
    AGENT_EXIT;
    return (unsigned char *)&long_ret;
}


int
write_dot1dStaticEntry(int action,
                         u_char * var_val,
                         u_char var_val_type,
                         size_t var_val_len,
                         u_char * statP, oid * name, size_t name_len)
{
    int i;
    int var = 0;
    int value = 0;
    static bridge_static_t *entry_row = NULL;

    /* Set the index */
    dot1dstatic_parms->index = static_entry_num;

    /* Find out which variable is being set */
    var = name[DOT1DSTATICENTRY_NAME_LENGTH -1];

    if (var_val) {
        value = * (int *) var_val;
    }

    /* 
     * First populate the entry either with existing values or the default
     * values
     */
    if (entry_row == NULL) {
        entry_row = (bridge_static_t *)sal_alloc(sizeof(bridge_static_t),
                                                 MIBII_DOT1DSTATIC_MIB);

        if (entry_row == NULL) {
            return SNMP_ERR_RESOURCEUNAVAILABLE;
        }

        /* Get the mac and port numbers from the oid */
        for (i = 0; i < MAC_ADDRESS_LEN; i++) {
            entry_row->mac[i] = (u_char)name[DOT1DSTATICENTRY_NAME_LENGTH+i];
            AGENT_DBG(("mac[%d] = %d\n", i, entry_row->mac[i]));
        }

        /* one based port */
        entry_row->port = 
                    (int)name[DOT1DSTATICENTRY_NAME_LENGTH + MAC_ADDRESS_LEN];
		AGENT_DBG(("Port number  = %d", entry_row->port));

        /* If not a pre-existing entry, set default values */
        if (statP == NULL) {
            /* Setting the vlan id to 1 */
            entry_row->isstatic = C_STATIC_STATUS_PERMANENT;
            entry_row->vid = 1;
        } else {
            /* existing entry, get the info using index */
            dot1dstatic_parms->group_obj = DOT1D_STATICDATA;
            dot1dstatic_parms->callback_rw = CALLBACK_ACT_READ;
            snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                                SNMPD_CALLBACK_BRIDGE,
                                dot1dstatic_parms);
            memcpy (entry_row,
                    &(dot1dstatic_parms->group_param.dot1dstat_param),
                                            sizeof(bridge_static_t));
        }
    }

    switch (action) {
    case RESERVE1:
        switch(var)
        {
            case DOT1DSTATICADDRESS:
                if (var_val_type != ASN_OCTET_STR) {
                    fprintf(stderr, "write to dot1dStatic not ASN_OCTET_STR\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len != sizeof(char) * MAC_ADDRESS_LEN) {
                    fprintf(stderr, "write to dot1dStatic: bad length\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGLENGTH;
                }
                break;
            case DOT1DSTATICRECEIVEPORT:
                if (var_val_type != ASN_INTEGER) {
                    fprintf(stderr, "write to dot1dStatic not ASN_INTEGER\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len > sizeof(long)) {
                    fprintf(stderr, "write to dot1dStatic: bad length\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGLENGTH;
                }

                if (value < 1 || value > interface_count) {
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGVALUE;
                }
                break;

            case DOT1DSTATICALLOWEDTOGOTO:
                if (var_val_type != ASN_OCTET_STR) {
                    fprintf(stderr, "write to dot1dStatic not ASN_OCTET_STR\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len != 
                        sizeof(char)*(interface_count/INTERFACE_BYTE_COUNT)) {
                    fprintf(stderr, "write to dot1dStatic: bad length\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGLENGTH;
                }
                break;
            case DOT1DSTATICSTATUS:
                if (var_val_type != ASN_INTEGER) {
                    fprintf(stderr, "write to dot1dStatic not ASN_INTEGER\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGTYPE;
                }
                if (var_val_len > sizeof(long)) {
                    fprintf(stderr, "write to dot1dStatic: bad length\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGLENGTH;
                }
                if (value < C_STATIC_STATUS_OTHER || 
                        value > C_STATIC_STATUS_DELONTMOUT) {
                    fprintf(stderr, "write to dot1dStatic: bad value\n");
                    AGENT_EXIT;
                    return SNMP_ERR_WRONGVALUE;
                }
                break;
        } /* switch(var) */
        break;

    case RESERVE2:
        break;

    case FREE:
        AGENT_DBG(("In free"));
        if (entry_row){
            sal_free(entry_row);
        }
        entry_row = NULL;
        break;

    case ACTION:
        break;

    case UNDO:
        if (entry_row) {
            sal_free (entry_row);
        }
        entry_row = NULL;
        break;

    case COMMIT:
        switch(var) {
            case DOT1DSTATICADDRESS:
                memcpy(entry_row->mac, var_val, var_val_len);
                break;

            case DOT1DSTATICRECEIVEPORT:
                entry_row->port = value;
                break;

            case DOT1DSTATICALLOWEDTOGOTO:
                /* We allow all ports....we set array to FF */
                memcpy (entry_row->mac, var_val, var_val_len);
                break;

            case DOT1DSTATICSTATUS:
                entry_row->isstatic = value;
                break;
        }

        /* We support (2), (3), (5) */
        dot1dstatic_parms->group_obj = DOT1D_STATICWRITE;
        dot1dstatic_parms->value = entry_row->isstatic;
        memcpy (&(dot1dstatic_parms->group_param.dot1dstat_param), entry_row,
                                                sizeof(bridge_static_t));
        dot1dstatic_parms->callback_rw = CALLBACK_ACT_WRITE;
        snmp_call_callbacks(SNMP_CALLBACK_APPLICATION, 
                            SNMPD_CALLBACK_BRIDGE,
                            dot1dstatic_parms);

        if (entry_row){
            sal_free(entry_row);
        }
        entry_row = NULL;

        if (dot1dstatic_parms->rv == MIB_STATUS_ERROR) {
            AGENT_EXIT;
            return SNMP_ERR_COMMITFAILED;
        }
        break;
    }

    AGENT_EXIT;
    return SNMP_ERR_NOERROR;
}

